"""
.. include:: ../README.md
"""

import numpy as np
import networkx as nx
from genice_core.topology import (
    noodlize,
    split_into_simple_paths,
    connect_matching_paths,
)
from genice_core.dipole import optimize, vector_sum, _dipole_moment_pbc
from typing import Union, List, Optional
from logging import getLogger, DEBUG


def ice_graph(
    g: nx.Graph,
    vertexPositions: Union[np.ndarray, None] = None,
    isPeriodicBoundary: bool = False,
    dipoleOptimizationCycles: int = 0,
    fixedEdges: nx.DiGraph = nx.DiGraph(),
    max_attempts: int = 100,
) -> Optional[nx.DiGraph]:
    """Make a digraph that obeys the ice rules.

    Args:
        g (nx.Graph): An ice-like undirected graph. Node labels in the graph g must be consecutive integers from 0 to N-1, where N is the number of nodes, and the labels correspond to the order in vertexPositions.
        vertexPositions (Union[nx.ndarray, None], optional): Positions of the vertices in N x 3 numpy array. Defaults to None.
        isPeriodicBoundary (bool, optional): If True, the positions are considered to be in the fractional coordinate system. Defaults to False.
        dipoleOptimizationCycles (int, optional): Number of iterations to reduce the net dipole moment. Defaults to 0 (no iteration).
        fixedEdges (nx.DiGraph, optional): A digraph made of edges whose directions are fixed. All edges in fixed must also be included in g. Defaults to an empty graph.
        max_attempts (int, optional): Maximum number of attempts to connect matching paths. Defaults to 100.

    Returns:
        Optional[nx.DiGraph]: An ice graph that obeys the ice rules, or None if no solution is found within max_attempts.
    """
    logger = getLogger()

    # derived cycles in extending the fixed edges.
    derivedCycles: List[List[int]] = []

    if fixedEdges.size() > 0:
        if logger.isEnabledFor(DEBUG):
            for edge in fixedEdges.edges():
                logger.debug(f"FIXED EDGE {edge}")

        # connect matching paths
        processedEdges = None
        for attempt in range(max_attempts):
            # It returns Nones when it fails to connect paths.
            # The processedEdges also include derivedCycles.
            processedEdges, derivedCycles = connect_matching_paths(fixedEdges, g)
            if processedEdges:
                break
            logger.info(f"Attempt {attempt + 1}/{max_attempts} failed to connect paths")
        else:
            logger.error(f"Failed to find a solution after {max_attempts} attempts")
            return None
    else:
        processedEdges = nx.DiGraph()

    # really fixed in connect_matching_paths()
    finallyFixedEdges = nx.DiGraph(processedEdges)
    for cycle in derivedCycles:
        for edge in zip(cycle, cycle[1:]):
            finallyFixedEdges.remove_edge(*edge)

    # Divide the remaining (unfixed) part of the graph into a noodle graph
    dividedGraph = noodlize(g, processedEdges)

    # Simplify paths ( paths with least crossings )
    paths = list(split_into_simple_paths(len(g), dividedGraph)) + derivedCycles

    # arrange the orientations here if you want to balance the polarization
    if vertexPositions is not None:
        # Set the targetPol in order to cancel the polarization in the fixed part.
        targetPol = -vector_sum(finallyFixedEdges, vertexPositions, isPeriodicBoundary)

        paths = optimize(
            paths,
            vertexPositions,
            isPeriodicBoundary=isPeriodicBoundary,
            dipoleOptimizationCycles=dipoleOptimizationCycles,
            targetPol=targetPol,
        )

    # Combine everything together
    dg = nx.DiGraph(finallyFixedEdges)

    for path in paths:
        nx.add_path(dg, path)

    # Verify that the graph obeys the ice rules
    for node in dg:
        if fixedEdges.has_node(node):
            if fixedEdges.in_degree(node) > 2 or fixedEdges.out_degree(node) > 2:
                continue
        assert (
            dg.in_degree(node) <= 2
        ), f"{node} {list(dg.successors(node))} {list(dg.predecessors(node))}"
        assert dg.out_degree(node) <= 2

    return dg



"""
Arrange edges appropriately.
"""

from logging import getLogger, DEBUG
import networkx as nx
import numpy as np
from typing import Union, List, Tuple, Optional


def _trace_path(g: nx.Graph, path: List[int]) -> List[int]:
    """Trace the path in a linear or cyclic graph.

    Args:
        g (nx.Graph): A linear or a simple cyclic graph.
        path (List[int]): A given path to be extended.

    Returns:
        List[int]: The extended path or cycle.
    """
    while True:
        # look at the head of the path
        last, head = path[-2:]
        for next_node in g[head]:
            if next_node != last:
                # go ahead
                break
        else:
            # no next node
            return path
        path.append(next_node)
        if next_node == path[0]:
            # is cyclic
            return path


def _find_path(g: nx.Graph) -> List[int]:
    """Find a path in a linear or cyclic graph.

    Args:
        g (nx.Graph): A linear or a simple cyclic graph.

    Returns:
        List[int]: The path or cycle.
    """
    nodes = list(g.nodes())
    # choose one node
    head = nodes[0]
    # look neighbors
    neighbors = list(g[head])
    if len(neighbors) == 0:
        # isolated node
        return []
    elif len(neighbors) == 1:
        # head is an end node, fortunately.
        return _trace_path(g, [head, neighbors[0]])
    # look forward
    c0 = _trace_path(g, [head, neighbors[0]])

    if c0[-1] == head:
        # cyclic graph
        return c0

    # look backward
    c1 = _trace_path(g, [head, neighbors[1]])
    return c0[::-1] + c1[1:]


def _divide(g: nx.Graph, vertex: int, offset: int) -> None:
    """Divide a vertex into two vertices and redistribute edges.

    Args:
        g (nx.Graph): The graph to modify.
        vertex (int): The vertex to divide.
        offset (int): The offset for the new vertex label.
    """
    # fill by Nones if number of neighbors is less than 4
    nei = (list(g[vertex]) + [None, None, None, None])[:4]

    # two neighbor nodes that are passed away to the new node
    migrants = set(np.random.choice(nei, 2, replace=False)) - {None}

    # new node label
    newVertex = vertex + offset

    # assemble edges
    for migrant in migrants:
        g.remove_edge(migrant, vertex)
        g.add_edge(newVertex, migrant)


def noodlize(g: nx.Graph, fixed: nx.DiGraph = nx.DiGraph()) -> nx.Graph:
    """Divide each vertex of the graph and make a set of paths.

    A new algorithm suggested by Prof. Sakuma, Yamagata University.

    Args:
        g (nx.Graph): An ice-like undirected graph. All vertices must not be >4-degree.
        fixed (nx.DiGraph, optional): Specifies the edges whose direction is fixed. Defaults to an empty graph.

    Returns:
        nx.Graph: A graph made of chains and cycles.
    """
    logger = getLogger()

    g_fix = nx.Graph(fixed)  # undirected copy

    offset = len(g)

    # divided graph
    g_noodles = nx.Graph(g)
    for edge in fixed.edges():
        g_noodles.remove_edge(*edge)

    for v in g:
        if g_fix.has_node(v):
            nfixed = g_fix.degree[v]
        else:
            nfixed = 0
        if nfixed == 0:
            _divide(g_noodles, v, offset)

    return g_noodles


def _decompose_complex_path(path: List[int]) -> List[List[int]]:
    """Divide a complex path with self-crossings into simple cycles and paths.

    Args:
        path (List[int]): A complex path.

    Yields:
        List[int]: A short and simple path/cycle.
    """
    logger = getLogger()
    if len(path) == 0:
        return
    logger.debug(f"decomposing {path}...")
    order = dict()
    order[path[0]] = 0
    store = [path[0]]
    headp = 1
    while headp < len(path):
        node = path[headp]

        if node in order:
            # it is a cycle!
            size = len(order) - order[node]
            cycle = store[-size:] + [node]
            yield cycle

            # remove them from the order[]
            for v in cycle[1:]:
                del order[v]

            # truncate the store
            store = store[:-size]

        order[node] = len(order)
        store.append(node)
        headp += 1
    if len(store) > 1:
        yield store
    logger.debug(f"Done decomposition.")


def split_into_simple_paths(
    nnode: int,
    g_noodles: nx.Graph,
) -> List[List[int]]:
    """Set the orientations to the components.

    Args:
        nnode (int): Number of nodes in the original graph.
        g_noodles (nx.Graph): The divided graph.

    Yields:
        List[int]: A short and simple path/cycle.
    """
    for verticeSet in nx.connected_components(g_noodles):
        # a component of c is either a chain or a cycle.
        g_noodle = g_noodles.subgraph(verticeSet)

        # Find a simple path in the doubled graph
        # It must be a simple path or a simple cycle.
        path = _find_path(g_noodle)

        # Flatten then path. It may make the path self-crossing.
        flatten = [v % nnode for v in path]

        # Divide a long path into simple paths and cycles.
        yield from _decompose_complex_path(flatten)


def _remove_dummy_nodes(g: Union[nx.Graph, nx.DiGraph]) -> None:
    """Remove dummy nodes from the graph.

    Args:
        g (Union[nx.Graph, nx.DiGraph]): The graph to clean.
    """
    for i in range(-1, -5, -1):
        if g.has_node(i):
            g.remove_node(i)


def connect_matching_paths(
    fixed: nx.DiGraph, g: nx.Graph
) -> Tuple[Optional[nx.DiGraph], List[List[int]]]:
    """Connect matching paths between two types of edges.

    This function creates a set of paths that connect edges of two different types (A and B).
    Each path connects one edge of type A with one edge of type B, and no two paths share any edges.
    The goal is to create a complete set of paths that covers all edges in the graph.

    Args:
        fixed (nx.DiGraph): Fixed edges.
        g (nx.Graph): Skeletal graph.

    Returns:
        Tuple[Optional[nx.DiGraph], List[List[int]]]: A tuple containing:
            - The extended fixed graph (derived cycles are included)
            - A list of derived cycles.
    """

    def _choose_free_edge(g: nx.Graph, dg: nx.DiGraph, node: int) -> Optional[int]:
        """Find an unfixed edge of the node.

        Args:
            g (nx.Graph): The original graph.
            dg (nx.DiGraph): The directed graph.
            node (int): The node to find edges for.

        Returns:
            Optional[int]: A free edge if found, None otherwise.
        """
        # add dummy nodes to make number of edges be four.
        neis = (list(g[node]) + [-1, -2, -3, -4])[:4]
        # and select one randomly
        np.random.shuffle(neis)
        for nei in neis:
            if not (dg.has_edge(node, nei) or dg.has_edge(nei, node)):
                return nei
        return None

    logger = getLogger()

    # Make a copy to keep the original graph untouched
    _fixed = nx.DiGraph(fixed)

    in_peri = set()
    out_peri = set()
    for node in _fixed:
        # If the node has unfixed edges,
        if _fixed.in_degree[node] + _fixed.out_degree[node] < g.degree[node]:
            # if it is not balanced,
            if _fixed.in_degree[node] > _fixed.out_degree[node]:
                out_peri.add(node)
            elif _fixed.in_degree[node] < _fixed.out_degree[node]:
                in_peri.add(node)

    logger.debug(f"out_peri {out_peri}")
    logger.debug(f"in_peri {in_peri}")

    derivedCycles = []

    while len(out_peri) > 0:
        node = np.random.choice(list(out_peri))
        out_peri -= {node}

        path = [node]
        while True:
            if node < 0:
                # Path search completed.
                logger.debug(f"Dead end at {node}. Path is {path}.")
                break
            if node in in_peri:
                # Path search completed.
                logger.debug(f"Reach at a perimeter node {node}. Path is {path}.")
                # in_peri and out_peri are now pair-annihilated.
                in_peri -= {node}
                break
            if node in out_peri:
                logger.debug(f"node {node} is on the out_peri...")
            # if the node can no longer be balanced,
            if max(_fixed.in_degree(node), _fixed.out_degree(node)) * 2 > 4:
                # Start over.
                logger.info(f"Failed to balance. Starting over ...")
                return None, None
            if g.degree(node) == _fixed.degree(node):
                # Start over.
                logger.info(f"node {node} has no free edge. Starting over ...")
                return None, None
            # Find the next node. That may be a decorated one.
            next = _choose_free_edge(g, _fixed, node)
            # fix the edge
            _fixed.add_edge(node, next)
            # record to the path
            if next >= 0:
                path.append(next)
                # if still incoming edges are more than outgoing ones,
                if _fixed.in_degree[node] > _fixed.out_degree[node]:
                    # It is still a perimeter.
                    out_peri.add(node)
            # go ahead
            node = next

            # if it is circular, i.e. if the last node of the path has already included in the path,
            try:
                loc = path[:-1].index(node)
                # Separate the cycle from the path and store in derivedCycles.
                derivedCycles.append(path[loc:])
                # and shorten the path
                path = path[: loc + 1]
            except ValueError:
                pass

    # starting from in_peri
    # Almost the same process, again.
    while len(in_peri) > 0:
        node = np.random.choice(list(in_peri))
        in_peri -= {node}
        logger.debug(
            f"first node {node}, its neighbors {g[node]} {list(_fixed.successors(node))} {list(_fixed.predecessors(node))}"
        )

        path = [node]
        while True:
            if node < 0:
                # Path search completed.
                logger.debug(f"Dead end at {node}. Path is {path} {in_peri}.")
                break
            if node in out_peri:
                # Path search completed.
                logger.debug(f"Reach at a perimeter node {node}. Path is {path}.")
                # in_peri and out_peri are now pair-annihilated.
                out_peri -= {node}
                break
            if node in in_peri:
                logger.debug(f"node {node} is on the in_peri...")
                # out_periのノードを何度も通ると、欠陥になってしまう。
            if max(_fixed.in_degree(node), _fixed.out_degree(node)) * 2 > 4:
                logger.info(f"Failed to balance. Starting over ...")
                return None, None
            if g.degree(node) == _fixed.degree(node):
                # Start over.
                logger.info(f"node {node} has no free edge. Starting over ...")
                return None, None
            next = _choose_free_edge(g, _fixed, node)
            # record to the path
            if next >= 0:
                path.append(next)
            # fix the edge  #####
            _fixed.add_edge(next, node)
            # if still incoming edges are more than outgoing ones,
            if next >= 0:
                #####
                if _fixed.in_degree[node] < _fixed.out_degree[node]:
                    in_peri.add(node)
                    logger.debug(
                        f"{node} is added to in_peri {_fixed.in_degree[node]} . {_fixed.out_degree[node]}"
                    )
            # go ahead
            node = next
            # if it is circular
            try:
                loc = path[:-1].index(node)
                derivedCycles.append(path[loc:])
                path = path[: loc + 1]
            except ValueError:
                pass

    if logger.isEnabledFor(DEBUG):
        logger.debug(f"size of g {g.number_of_edges()}")
        logger.debug(f"size of fixed {_fixed.number_of_edges()}")
        assert len(in_peri) == 0, f"In-peri remains. {in_peri}"
        assert len(out_peri) == 0, f"Out-peri remains. {out_peri}"
        logger.debug("re-check perimeters")

        in_peri = set()
        out_peri = set()
        for node in _fixed:
            if node >= 0:
                if _fixed.in_degree[node] + _fixed.out_degree[node] < g.degree[node]:
                    if _fixed.in_degree[node] > _fixed.out_degree[node]:
                        out_peri.add(node)
                    elif _fixed.in_degree[node] < _fixed.out_degree[node]:
                        in_peri.add(node)

        assert len(in_peri) == 0, f"In-peri remains. {in_peri}"
        assert len(out_peri) == 0, f"Out-peri remains. {out_peri}"

        # 拡大したグラフが指定された固定辺をすべて含んでいることを確認。
        for edge in fixed.edges():
            assert _fixed.has_edge(*edge)

    # # remove edges in derivedCycles from _fixed
    # for cycle in derivedCycles:
    #     for edge in zip(cycle, cycle[1:]):
    #         _fixed.remove_edge(*edge)

    _remove_dummy_nodes(_fixed)

    if logger.isEnabledFor(DEBUG):
        logger.debug(f"Number of fixed edges is {_fixed.size()} / {g.size()}")
        logger.debug(f"Number of free cycles: {len(derivedCycles)}")
        ne = sum([len(cycle) - 1 for cycle in derivedCycles])
        logger.debug(f"Number of edges in free cycles: {ne}")

    return _fixed, derivedCycles



"""
Optimizes the orientations of directed paths to reduce the net dipole moment.
"""

from logging import getLogger, DEBUG
from typing import Union, List, Optional
import math
import random

import numpy as np
import networkx as nx


def vector_sum(
    dg: nx.DiGraph, vertexPositions: np.ndarray, isPeriodicBoundary: bool = False
) -> np.ndarray:
    """Calculate the net polarization (vector sum) of a digraph.

    Args:
        dg (nx.DiGraph): The digraph.
        vertexPositions (np.ndarray): Positions of the vertices.
        isPeriodicBoundary (bool, optional): If true, the vertex positions must be in fractional coordinate. Defaults to False.

    Returns:
        np.ndarray: Net polarization vector.
    """
    pol = np.zeros_like(vertexPositions[0])
    for i, j in dg.edges():
        d = vertexPositions[j] - vertexPositions[i]
        if isPeriodicBoundary:
            d -= np.floor(d + 0.5)
        pol += d
    return pol


def _dipole_moment_pbc(path: List[int], vertexPositions: np.ndarray) -> np.ndarray:
    """Calculate the dipole moment of a path with periodic boundary conditions.

    Args:
        path (List[int]): The path to calculate dipole moment for.
        vertexPositions (np.ndarray): Positions of the vertices.

    Returns:
        np.ndarray: The dipole moment vector.
    """
    # vectors between adjacent vertices.
    relativeVector = vertexPositions[path[1:]] - vertexPositions[path[:-1]]
    # PBC wrap
    relativeVector -= np.floor(relativeVector + 0.5)
    # total dipole along the chain (or a cycle)
    return np.sum(relativeVector, axis=0)


def optimize(
    paths: List[List[int]],
    vertexPositions: np.ndarray,
    dipoleOptimizationCycles: int = 2000,
    isPeriodicBoundary: bool = False,
    targetPol: Optional[np.ndarray] = None,
) -> List[List[int]]:
    """Minimize the net polarization by flipping several paths using simulated annealing.

    It is assumed that every vector has an identical dipole moment.

    Args:
        paths (List[List[int]]): List of directed paths. A path is a list of integers. A path with identical labels at first and last items are considered to be cyclic.
        vertexPositions (np.ndarray): Positions of the nodes.
        dipoleOptimizationCycles (int, optional): Number of iterations for the optimization. Defaults to 2000.
        isPeriodicBoundary (bool, optional): If `True`, the positions of the nodes must be in the fractional coordinate system. Defaults to False.
        targetPol (Optional[np.ndarray], optional): Target value for the dipole-moment optimization. Defaults to None.

    Returns:
        List[List[int]]: Optimized paths with minimized net polarization.
    """
    logger = getLogger()

    if targetPol is None:
        targetPol = np.zeros_like(vertexPositions[0])

    # polarized chains and cycles. Small cycle of dipoles are eliminated.
    polarizedEdges: List[int] = []

    dipoles: List[np.ndarray] = []
    for i, path in enumerate(paths):
        if isPeriodicBoundary:
            chainPol = _dipole_moment_pbc(path, vertexPositions)
            # if it is large enough, i.e. if it is a spanning cycle or a chain
            if chainPol @ chainPol > 1e-6:
                dipoles.append(chainPol)
                polarizedEdges.append(i)
        else:
            # dipole moment of a path; NOTE: No PBC.
            if path[0] != path[-1]:
                # If no PBC, a chain pol is simply an end-to-end pol.
                chainPol = vertexPositions[path[-1]] - vertexPositions[path[0]]
                dipoles.append(chainPol)
                polarizedEdges.append(i)
    dipoles = np.array(dipoles)

    if len(dipoles) == 0:
        return paths

    # Initialize simulated annealing parameters
    initial_temp = 1.0
    final_temp = 1e-4
    cooling_rate = (final_temp / initial_temp) ** (1.0 / dipoleOptimizationCycles)
    current_temp = initial_temp

    # Initialize current solution
    current_parities = np.random.randint(2, size=len(dipoles)) * 2 - 1
    current_pol = current_parities @ dipoles - targetPol
    best_parities = current_parities.copy()
    best_pol = current_pol.copy()
    no_improvement_count = 0
    last_improvement = 0

    if logger.isEnabledFor(DEBUG):
        logger.debug(f"initial {current_pol} target {targetPol}")
        logger.debug(f"dipoles {dipoles}")
        for i, parity in zip(polarizedEdges, current_parities):
            logger.debug(f"{parity}: {paths[i]}")

    for loop in range(dipoleOptimizationCycles):
        # Generate new solution by flipping a random subset of paths
        new_parities = current_parities.copy()
        num_flips = max(1, int(len(dipoles) * 0.1))  # Flip 10% of paths
        flip_indices = random.sample(range(len(dipoles)), num_flips)
        new_parities[flip_indices] *= -1

        # Calculate new polarization
        new_pol = new_parities @ dipoles - targetPol

        # Calculate energy difference
        delta_e = new_pol @ new_pol - current_pol @ current_pol

        # Accept new solution based on temperature and energy difference
        if delta_e < 0 or random.random() < math.exp(-delta_e / current_temp):
            current_parities = new_parities
            current_pol = new_pol

            # Update best solution if better
            if current_pol @ current_pol < best_pol @ best_pol:
                best_parities = current_parities.copy()
                best_pol = current_pol.copy()
                no_improvement_count = 0
                last_improvement = loop
                logger.debug(f"Depol. loop {loop}: {best_pol}")
            else:
                no_improvement_count += 1

        # Cool down
        current_temp *= cooling_rate

        # Early stopping if no improvement for a while
        if no_improvement_count > dipoleOptimizationCycles // 10:
            logger.debug(
                f"Early stopping at loop {loop} (no improvement since loop {last_improvement})"
            )
            break

        # Check convergence
        if best_pol @ best_pol < 1e-4:
            logger.debug("Optimized.")
            break

    logger.info(f"Depol. loop {loop}: {best_pol}")

    # Apply the best solution found
    for i, parity in zip(polarizedEdges, best_parities):
        if parity < 0:
            # invert the chain
            paths[i] = paths[i][::-1]

    return paths



"""
Arrange edges appropriately.

どうもバグを入れてしまったらしい。結果が変わる。

"""

from logging import getLogger, DEBUG
import networkx as nx
import numpy as np
from typing import Union, List, Tuple, Optional


def _trace_path(g: nx.Graph, path: List[int]) -> List[int]:
    """Trace the path in a linear or cyclic graph.

    Args:
        g (nx.Graph): A linear or a simple cyclic graph.
        path (List[int]): A given path to be extended.

    Returns:
        List[int]: The extended path or cycle.
    """
    while True:
        # look at the head of the path
        last, head = path[-2:]
        for next_node in g[head]:
            if next_node != last:
                # go ahead
                break
        else:
            # no next node
            return path
        path.append(next_node)
        if next_node == path[0]:
            # is cyclic
            return path


def _find_path(g: nx.Graph) -> List[int]:
    """Find a path in a linear or cyclic graph.

    Args:
        g (nx.Graph): A linear or a simple cyclic graph.

    Returns:
        List[int]: The path or cycle.
    """
    nodes = list(g.nodes())
    # choose one node
    head = nodes[0]
    # look neighbors
    neighbors = list(g[head])
    if len(neighbors) == 0:
        # isolated node
        return []
    elif len(neighbors) == 1:
        # head is an end node, fortunately.
        return _trace_path(g, [head, neighbors[0]])
    # look forward
    c0 = _trace_path(g, [head, neighbors[0]])

    if c0[-1] == head:
        # cyclic graph
        return c0

    # look backward
    c1 = _trace_path(g, [head, neighbors[1]])
    return c0[::-1] + c1[1:]


def _divide(g: nx.Graph, vertex: int, offset: int) -> None:
    """Divide a vertex into two vertices and redistribute edges.

    Args:
        g (nx.Graph): The graph to modify.
        vertex (int): The vertex to divide.
        offset (int): The offset for the new vertex label.
    """
    # fill by Nones if number of neighbors is less than 4
    nei = (list(g[vertex]) + [None, None, None, None])[:4]

    # two neighbor nodes that are passed away to the new node
    migrants = set(np.random.choice(nei, 2, replace=False)) - {None}

    # new node label
    newVertex = vertex + offset

    # assemble edges
    for migrant in migrants:
        g.remove_edge(migrant, vertex)
        g.add_edge(newVertex, migrant)


def noodlize(g: nx.Graph, fixed: nx.DiGraph = nx.DiGraph()) -> nx.Graph:
    """Divide each vertex of the graph and make a set of paths.

    A new algorithm suggested by Prof. Sakuma, Yamagata University.
    The vertices are divided based on their degrees and connectivity to minimize
    the number of crossings in the resulting paths.

    Args:
        g (nx.Graph): An ice-like undirected graph. All vertices must not be >4-degree.
        fixed (nx.DiGraph, optional): Specifies the edges whose direction is fixed. Defaults to an empty graph.

    Returns:
        nx.Graph: A graph made of chains and cycles.
    """
    logger = getLogger()

    g_fix = nx.Graph(fixed)  # undirected copy

    offset = len(g)

    # divided graph
    g_noodles = nx.Graph(g)
    for edge in fixed.edges():
        g_noodles.remove_edge(*edge)

    # Sort vertices by degree in descending order
    vertices = sorted(g.nodes(), key=lambda x: g.degree(x), reverse=True)

    for v in vertices:
        if g_fix.has_node(v):
            nfixed = g_fix.degree[v]
        else:
            nfixed = 0
        if nfixed == 0:
            # Choose neighbors to migrate based on their degrees
            neighbors = list(g_noodles[v])
            if len(neighbors) > 2:
                # Prefer neighbors with higher degrees to minimize crossings
                neighbors.sort(key=lambda x: g_noodles.degree(x), reverse=True)
                # Choose the two neighbors with highest degrees
                migrants = set(neighbors[:2])
                _divide(g_noodles, v, offset)

    return g_noodles


def _decompose_complex_path(path: List[int]) -> List[List[int]]:
    """Divide a complex path with self-crossings into simple cycles and paths.

    Args:
        path (List[int]): A complex path.

    Yields:
        List[int]: A short and simple path/cycle.
    """
    logger = getLogger()
    if len(path) == 0:
        return
    logger.debug(f"decomposing {path}...")
    order = dict()
    order[path[0]] = 0
    store = [path[0]]
    headp = 1
    while headp < len(path):
        node = path[headp]

        if node in order:
            # it is a cycle!
            size = len(order) - order[node]
            cycle = store[-size:] + [node]
            yield cycle

            # remove them from the order[]
            for v in cycle[1:]:
                del order[v]

            # truncate the store
            store = store[:-size]

        order[node] = len(order)
        store.append(node)
        headp += 1
    if len(store) > 1:
        yield store
    logger.debug(f"Done decomposition.")


def split_into_simple_paths(
    nnode: int,
    g_noodles: nx.Graph,
) -> List[List[int]]:
    """Set the orientations to the components.

    Args:
        nnode (int): Number of nodes in the original graph.
        g_noodles (nx.Graph): The divided graph.

    Yields:
        List[int]: A short and simple path/cycle.
    """
    for verticeSet in nx.connected_components(g_noodles):
        # a component of c is either a chain or a cycle.
        g_noodle = g_noodles.subgraph(verticeSet)

        # Find a simple path in the doubled graph
        # It must be a simple path or a simple cycle.
        path = _find_path(g_noodle)

        # Flatten then path. It may make the path self-crossing.
        flatten = [v % nnode for v in path]

        # Divide a long path into simple paths and cycles.
        yield from _decompose_complex_path(flatten)


def _remove_dummy_nodes(g: Union[nx.Graph, nx.DiGraph]) -> None:
    """Remove dummy nodes from the graph.

    Args:
        g (Union[nx.Graph, nx.DiGraph]): The graph to clean.
    """
    for i in range(-1, -5, -1):
        if g.has_node(i):
            g.remove_node(i)


def balance(
    fixed: nx.DiGraph, g: nx.Graph
) -> Tuple[Optional[nx.DiGraph], List[List[int]]]:
    """Extend the prefixed digraph to make the remaining graph balanced.

    Args:
        fixed (nx.DiGraph): Fixed edges.
        g (nx.Graph): Skeletal graph.

    Returns:
        Tuple[Optional[nx.DiGraph], List[List[int]]]: A tuple containing:
            - The extended fixed graph (derived cycles are included)
            - A list of derived cycles.
    """

    def _choose_free_edge(g: nx.Graph, dg: nx.DiGraph, node: int) -> Optional[int]:
        """Find an unfixed edge of the node.

        Args:
            g (nx.Graph): The original graph.
            dg (nx.DiGraph): The directed graph.
            node (int): The node to find edges for.

        Returns:
            Optional[int]: A free edge if found, None otherwise.
        """
        # add dummy nodes to make number of edges be four.
        neis = (list(g[node]) + [-1, -2, -3, -4])[:4]
        # and select one randomly
        np.random.shuffle(neis)
        for nei in neis:
            if not (dg.has_edge(node, nei) or dg.has_edge(nei, node)):
                return nei
        return None

    def _try_balance(
        _fixed: nx.DiGraph,
        in_peri: set,
        out_peri: set,
        derived_cycles: List[List[int]],
        max_attempts: int = 100,
    ) -> Tuple[Optional[nx.DiGraph], List[List[int]]]:
        """Try to balance the graph with backtracking.

        Args:
            _fixed (nx.DiGraph): The current fixed edges.
            in_peri (set): Set of nodes with more incoming edges.
            out_peri (set): Set of nodes with more outgoing edges.
            derived_cycles (List[List[int]]): List of derived cycles.
            max_attempts (int, optional): Maximum number of attempts. Defaults to 100.

        Returns:
            Tuple[Optional[nx.DiGraph], List[List[int]]]: The balanced graph and derived cycles if successful, None otherwise.
        """
        if len(out_peri) == 0 and len(in_peri) == 0:
            return _fixed, derived_cycles

        if max_attempts <= 0:
            return None, None

        # Try to balance from out_peri first
        if len(out_peri) > 0:
            node = np.random.choice(list(out_peri))
            out_peri -= {node}
            path = [node]
            while True:
                if node < 0:
                    logger.debug(f"Dead end at {node}. Path is {path}.")
                    break
                if node in in_peri:
                    logger.debug(f"Reach at a perimeter node {node}. Path is {path}.")
                    in_peri -= {node}
                    break
                if node in out_peri:
                    logger.debug(f"node {node} is on the out_peri...")
                if max(_fixed.in_degree(node), _fixed.out_degree(node)) * 2 > 4:
                    break
                next_node = _choose_free_edge(g, _fixed, node)
                if next_node is None:
                    break
                _fixed.add_edge(node, next_node)
                if next_node >= 0:
                    path.append(next_node)
                    if _fixed.in_degree(node) > _fixed.out_degree(node):
                        out_peri.add(node)
                node = next_node
                try:
                    loc = path[:-1].index(node)
                    derived_cycles.append(path[loc:])
                    path = path[: loc + 1]
                except ValueError:
                    pass
            # Try to balance the remaining graph
            result = _try_balance(
                _fixed, in_peri, out_peri, derived_cycles, max_attempts - 1
            )
            if result[0] is not None:
                return result
        # If out_peri balancing failed, try in_peri
        if len(in_peri) > 0:
            node = np.random.choice(list(in_peri))
            in_peri -= {node}
            logger.debug(
                f"first node {node}, its neighbors {g[node]} {list(_fixed.successors(node))} {list(_fixed.predecessors(node))}"
            )
            path = [node]
            while True:
                if node < 0:
                    logger.debug(f"Dead end at {node}. Path is {path} {in_peri}.")
                    break
                if node in out_peri:
                    logger.debug(f"Reach at a perimeter node {node}. Path is {path}.")
                    out_peri -= {node}
                    break
                if node in in_peri:
                    logger.debug(f"node {node} is on the in_peri...")
                if max(_fixed.in_degree(node), _fixed.out_degree(node)) * 2 > 4:
                    break
                next_node = _choose_free_edge(g, _fixed, node)
                if next_node is None:
                    break
                if next_node >= 0:
                    path.append(next_node)
                _fixed.add_edge(next_node, node)
                if next_node >= 0:
                    if _fixed.in_degree(node) < _fixed.out_degree(node):
                        in_peri.add(node)
                        logger.debug(
                            f"{node} is added to in_peri {_fixed.in_degree[node]} . {_fixed.out_degree[node]}"
                        )
                node = next_node
                try:
                    loc = path[:-1].index(node)
                    derived_cycles.append(path[loc:])
                    path = path[: loc + 1]
                except ValueError:
                    pass
            # Try to balance the remaining graph
            result = _try_balance(
                _fixed, in_peri, out_peri, derived_cycles, max_attempts - 1
            )
            if result[0] is not None:
                return result

        return None, None

    logger = getLogger()

    # Make a copy to keep the original graph untouched
    _fixed = nx.DiGraph(fixed)

    in_peri = set()
    out_peri = set()
    for node in _fixed:
        # If the node has unfixed edges,
        if _fixed.in_degree[node] + _fixed.out_degree[node] < g.degree[node]:
            # if it is not balanced,
            if _fixed.in_degree[node] > _fixed.out_degree[node]:
                out_peri.add(node)
            elif _fixed.in_degree[node] < _fixed.out_degree[node]:
                in_peri.add(node)

    logger.debug(f"out_peri {out_peri}")
    logger.debug(f"in_peri {in_peri}")

    derived_cycles: List[List[int]] = []
    result = _try_balance(_fixed, in_peri, out_peri, derived_cycles)

    if result[0] is None:
        logger.info("Failed to balance the graph.")
        return None, None

    _fixed, derived_cycles = result

    if logger.isEnabledFor(DEBUG):
        logger.debug(f"size of g {g.number_of_edges()}")
        logger.debug(f"size of fixed {_fixed.number_of_edges()}")
        assert len(in_peri) == 0, f"In-peri remains. {in_peri}"
        assert len(out_peri) == 0, f"Out-peri remains. {out_peri}"
        logger.debug("re-check perimeters")

        in_peri = set()
        out_peri = set()
        for node in _fixed:
            if node >= 0:
                if _fixed.in_degree[node] + _fixed.out_degree[node] < g.degree[node]:
                    if _fixed.in_degree[node] > _fixed.out_degree[node]:
                        out_peri.add(node)
                    elif _fixed.in_degree[node] < _fixed.out_degree[node]:
                        in_peri.add(node)

        assert len(in_peri) == 0, f"In-peri remains. {in_peri}"
        assert len(out_peri) == 0, f"Out-peri remains. {out_peri}"

        # Verify that the extended graph contains all fixed edges
        for edge in fixed.edges():
            assert _fixed.has_edge(*edge)

    _remove_dummy_nodes(_fixed)

    if logger.isEnabledFor(DEBUG):
        logger.debug(f"Number of fixed edges is {_fixed.size()} / {g.size()}")
        logger.debug(f"Number of free cycles: {len(derived_cycles)}")
        ne = sum([len(cycle) - 1 for cycle in derived_cycles])
        logger.debug(f"Number of edges in free cycles: {ne}")

    return _fixed, derived_cycles
